defmodule Nulla.Relations do
  alias Nulla.Actors.Actor

  @moduledoc """
  The Relations context.
  """

  import Ecto.Query, warn: false
  alias Nulla.Repo

  alias Nulla.Relations.Relation

  @doc """
  Returns the list of relations.

  ## Examples

      iex> list_relations()
      [%Relation{}, ...]

  """
  def list_relations do
    Repo.all(Relation)
  end

  @doc """
  Gets a single relation.

  Raises `Ecto.NoResultsError` if the Relation does not exist.

  ## Examples

      iex> get_relation!(123)
      %Relation{}

      iex> get_relation!(456)
      ** (Ecto.NoResultsError)

  """
  def get_relation!(id), do: Repo.get!(Relation, id)

  def get_relation_by(by) when is_map(by) or is_list(by), do: Repo.get_by(Relation, by)

  @doc """
  Creates a relation.

  ## Examples

      iex> create_relation(%{field: value})
      {:ok, %Relation{}}

      iex> create_relation(%{field: bad_value})
      {:error, %Ecto.Changeset{}}

  """
  def create_relation(attrs \\ %{}) do
    %Relation{}
    |> Relation.changeset(attrs)
    |> Repo.insert()
  end

  @doc """
  Updates a relation.

  ## Examples

      iex> update_relation(relation, %{field: new_value})
      {:ok, %Relation{}}

      iex> update_relation(relation, %{field: bad_value})
      {:error, %Ecto.Changeset{}}

  """
  def update_relation(%Relation{} = relation, attrs) do
    relation
    |> Relation.changeset(attrs)
    |> Repo.update()
  end

  @doc """
  Deletes a relation.

  ## Examples

      iex> delete_relation(relation)
      {:ok, %Relation{}}

      iex> delete_relation(relation)
      {:error, %Ecto.Changeset{}}

  """
  def delete_relation(%Relation{} = relation) do
    Repo.delete(relation)
  end

  @doc """
  Returns an `%Ecto.Changeset{}` for tracking relation changes.

  ## Examples

      iex> change_relation(relation)
      %Ecto.Changeset{data: %Relation{}}

  """
  def change_relation(%Relation{} = relation, attrs \\ %{}) do
    Relation.changeset(relation, attrs)
  end

  def count_following(local_actor_id) do
    Relation
    |> where([r], r.local_actor_id == ^local_actor_id and r.following == true)
    |> select([r], count(r.id))
    |> Repo.one()
  end

  def get_following(local_actor_id, page, limit) when is_integer(page) and page > 0 do
    offset = (page - 1) * limit

    query =
      from r in Relation,
        join: a in Actor,
        on: a.id == r.remote_actor_id,
        where: r.local_actor_id == ^local_actor_id and r.following == true,
        order_by: [asc: a.published],
        offset: ^offset,
        limit: ^limit,
        select: a

    Repo.all(query)
  end

  def count_followers(local_actor_id) do
    Relation
    |> where([r], r.local_actor_id == ^local_actor_id and r.followed_by == true)
    |> select([r], count(r.id))
    |> Repo.one()
  end

  def get_followers(local_actor_id, page, limit) when is_integer(page) and page > 0 do
    offset = (page - 1) * limit

    query =
      from r in Relation,
        join: a in Actor,
        on: a.id == r.remote_actor_id,
        where: r.local_actor_id == ^local_actor_id and r.followed_by == true,
        order_by: [asc: a.published],
        offset: ^offset,
        limit: ^limit,
        select: a

    Repo.all(query)
  end
end
